"""Factories for the ``conversation`` app."""
import factory

from django_libs.tests.factories import UserFactory

from ..models import Conversation, Message


class ConversationFactory(factory.DjangoModelFactory):
    FACTORY_FOR = Conversation

    @factory.post_generation
    def groups(self, create, extracted, **kwargs):
        self.users.add(UserFactory())


class MessageFactory(factory.DjangoModelFactory):
    FACTORY_FOR = Message

    user = factory.SubFactory(UserFactory)
    conversation = factory.SubFactory(ConversationFactory)
    text = factory.Sequence(lambda n: 'This is a text {}'.format(n))



"""Tests for the forms of the ``conversation`` app."""
from django.test import TestCase

from django_libs.tests.factories import UserFactory

from .. import forms
from ..models import Conversation


class MessageFormTestCase(TestCase):
    """Tests for the ``MessageForm`` model."""
    longMessage = True

    def setUp(self):
        self.user = UserFactory()
        self.other_user = UserFactory()
        self.content_object = UserFactory()

    def test_form(self):
        data = {'text': 'Foo'}
        form = forms.MessageForm(user=self.user, conversation=None, data=data,
                                 initial_user=None, content_object=None)
        self.assertTrue(form.errors)
        self.assertFalse(form.is_valid())
        data.update({'recipients': ['999']})
        form = forms.MessageForm(user=self.user, conversation=None, data=data,
                                 initial_user=None, content_object=None)
        self.assertTrue(form.errors)
        self.assertFalse(form.is_valid())
        data.update({'recipients': ['{}'.format(self.other_user.pk)]})
        form = forms.MessageForm(user=self.user, conversation=None, data=data,
                                 initial_user=None, content_object=None)
        self.assertFalse(form.errors)
        self.assertTrue(form.is_valid())
        conversation = form.save()
        self.assertEqual(Conversation.objects.count(), 1, msg=(
            'A new conversation should\'ve been started with the message.'))
        form = forms.MessageForm(user=self.user, data=data, conversation=None,
                                 initial_user=None, content_object=None)
        form.cleaned_data = data
        form.save()
        self.assertEqual(Conversation.objects.count(), 1, msg=(
            'The existing conversation should\'ve been re-used.'))
        form = forms.MessageForm(user=self.user, data=data, initial_user=None,
                                 conversation=conversation,
                                 content_object=None)
        form.save()
        self.assertEqual(Conversation.objects.count(), 1, msg=(
            'The existing conversation should\'ve been re-used.'))
        form = forms.MessageForm(user=self.user, conversation=None, data=data,
                                 initial_user=None,
                                 content_object=self.content_object)
        self.assertFalse(form.errors)
        self.assertTrue(form.is_valid())
        conversation = form.save()
        self.assertEqual(Conversation.objects.count(), 2, msg=(
            'A new conversation with a related object should\'ve been started'
            ' with the message.'))
        form = forms.MessageForm(user=self.user, data=data, initial_user=None,
                                 conversation=conversation,
                                 content_object=self.content_object)
        form.save()
        self.assertEqual(Conversation.objects.count(), 2, msg=(
            'The existing conversation should\'ve been re-used.'))



"""View mixins for the ``conversation`` app views."""
from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.core.urlresolvers import reverse
from django.http import Http404, HttpResponse
from django.utils.decorators import method_decorator

from django_libs.loaders import load_member

from .forms import MessageForm
from .models import Conversation


class ConversationViewMixin(object):
    """Mixin for conversation related views."""
    model = Conversation

    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        self.user = request.user
        if kwargs.get('user_pk'):
            try:
                self.initial_user = User.objects.get(pk=kwargs['user_pk'])
            except User.DoesNotExist:
                raise Http404
        if kwargs.get('c_type') and kwargs.get('obj_id'):
            try:
                content_type = ContentType.objects.get(model=kwargs['c_type'])
            except ContentType.DoesNotExist:
                raise Http404
            try:
                self.content_object = content_type.get_object_for_this_type(
                    pk=kwargs['obj_id'])
            except ObjectDoesNotExist:
                raise Http404
        if kwargs.get('pk'):
            # If it's not a CreateView, check the permission
            self.kwargs = kwargs
            self.object = self.get_object()
            if not request.user in self.object.users.all():
                raise Http404
        return super(ConversationViewMixin, self).dispatch(
            request, *args, **kwargs)

    def get_form_class(self):
        if hasattr(settings, 'CONVERSATION_MESSAGE_FORM'):
            return load_member(settings.CONVERSATION_MESSAGE_FORM)
        return MessageForm

    def get_form_kwargs(self, *args, **kwargs):
        kwargs = super(ConversationViewMixin, self).get_form_kwargs(
            *args, **kwargs)
        kwargs.update({
            'user': self.user,
            'conversation': self.object,
            'instance': None,
            'initial_user': self.initial_user if hasattr(
                self, 'initial_user') else None,
            'content_object': self.content_object if hasattr(
                self, 'content_object') else None,
        })
        return kwargs

    def get_context_data(self, **kwargs):
        context = super(ConversationViewMixin, self).get_context_data(**kwargs)
        if hasattr(self, 'initial_user'):
            context.update({'initial_user': self.initial_user})
        if self.object:
            context.update({'content_object': self.object.content_object})
        elif hasattr(self, 'content_object'):
            context.update({'content_object': self.content_object})
        return context

    def get_success_url(self):
        return reverse('conversation_update', kwargs={'pk': self.object.pk})


class ConversationStatusViewMixin(object):
    """
    Mixin for conversation related views, which change the conversation's
    status.

    """
    model = Conversation

    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        if request.method == 'POST' and request.is_ajax():
            self.kwargs = kwargs
            self.object = self.get_object()
            if not request.user in self.object.users.all():
                raise Http404
            if self.action == 'archive':
                self.object.archived_by.add(request.user)
                return HttpResponse('archived')
        raise Http404



"""Models for the conversation app."""
from django.contrib.contenttypes import generic
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.translation import ugettext_lazy as _


class Conversation(models.Model):
    """
    Model to contain different messages between one or more users.

    :users: Users participating in this conversation.
    :archived_by: List of participants, who archived this conversation.
    :read_by: List of participants, who read this conversation.
    :content_object: Optional related object the users are talking about.

    """
    users = models.ManyToManyField(
        'auth.User',
        verbose_name=_('Users'),
        related_name='conversations',
    )

    archived_by = models.ManyToManyField(
        'auth.User',
        verbose_name=_('Archived by'),
        related_name='archived_conversations',
    )

    read_by = models.ManyToManyField(
        'auth.User',
        verbose_name=_('Read by'),
        related_name='read_conversations',
    )

    # Generic FK to the object this conversation is about
    content_type = models.ForeignKey(
        ContentType,
        related_name='conversation_content_objects',
        null=True, blank=True,
    )
    object_id = models.PositiveIntegerField(null=True, blank=True)
    content_object = generic.GenericForeignKey('content_type', 'object_id')


class Message(models.Model):
    """
    Model, which holds information about a post within one conversation.

    :user: User, who posted the message.
    :conversation: Conversation, which contains this message.
    :date: Date the message was posted.
    :text: Message text.

    """
    user = models.ForeignKey(
        'auth.User',
        verbose_name=_('User'),
        related_name='messages',
    )

    conversation = models.ForeignKey(
        Conversation,
        verbose_name=_('Conversation'),
        related_name='messages',
    )

    date = models.DateTimeField(
        auto_now_add=True,
        verbose_name=_('Date'),
    )

    text = models.TextField(
        max_length=2048,
        verbose_name=_('Text'),
    )



"""Forms for the ``conversation`` app."""
from django import forms
from django.contrib.contenttypes.models import ContentType
from django.contrib.auth.models import User
from django.db.models import Q

from .models import Conversation, Message


class MessageForm(forms.ModelForm):
    """Form to post new messages to a new or existing conversation."""
    def __init__(self, user, conversation, initial_user, content_object, *args,
                 **kwargs):
        self.user = user
        self.conversation = conversation
        self.content_object = content_object
        super(MessageForm, self).__init__(*args, **kwargs)
        if not self.conversation:
            users = User.objects.exclude(pk=user.pk)
            self.fields['recipients'] = forms.MultipleChoiceField(
                choices=[(u.pk, u) for u in users],
                initial=([initial_user.pk] if initial_user else ''),
            )
            self.fields['recipients'].hidden = True

    def save(self, *args, **kwargs):
        if not self.instance.pk:
            self.instance.user = self.user
            if not self.conversation:
                # Check for existing conversations of these users
                recipients = User.objects.filter(
                    Q(pk__in=self.cleaned_data['recipients'])
                    | Q(pk=self.user.pk),
                )
                conversations = self.user.conversations.filter(
                    users__in=recipients)
                if self.content_object:
                    conversations = conversations.filter(
                        content_type=ContentType.objects.get_for_model(
                            self.content_object),
                        object_id=self.content_object.pk,
                    )
                if conversations:
                    self.conversation = conversations[0]
                else:
                    if self.content_object:
                        self.conversation = Conversation.objects.create(
                            content_object=self.content_object)
                    else:
                        self.conversation = Conversation.objects.create()
                    self.conversation.users.add(self.user)
                    for user in recipients:
                        self.conversation.users.add(user)
            self.instance.conversation = self.conversation

            # Reset archive marks
            self.instance.conversation.archived_by.clear()
            # Reset reading status
            self.instance.conversation.read_by.clear()
        return super(MessageForm, self).save(*args, **kwargs).conversation

    class Meta:
        model = Message
        fields = ('text', )



"""Tests for the models of the ``conversation`` app."""
from django.test import TestCase

from .factories import ConversationFactory, MessageFactory


class ConversationTestCase(TestCase):
    """Tests for the ``Conversation`` model."""
    longMessage = True

    def setUp(self):
        self.conversation = ConversationFactory()

    def test_model(self):
        self.assertTrue(self.conversation.pk, msg=(
            'Should be able to instantiate and save the object.'))


class MessageTestCase(TestCase):
    """Tests for the ``Message`` model."""
    longMessage = True

    def setUp(self):
        self.message = MessageFactory()

    def test_model(self):
        self.assertTrue(self.message.pk, msg=(
            'Should be able to instantiate and save the object.'))



"""Views for the ``conversation`` app."""
from django.core.urlresolvers import reverse
from django.views.generic import (
    CreateView,
    DetailView,
    RedirectView,
    UpdateView,
)

from . import view_mixins


class ConversationUpdateView(view_mixins.ConversationViewMixin, UpdateView):
    """View to update a conversation."""
    def dispatch(self, request, *args, **kwargs):
        resp = super(ConversationUpdateView, self).dispatch(
            request, *args, **kwargs)
        if hasattr(self, 'object'):
            self.object.read_by.add(request.user)
        return resp


class ConversationCreateView(view_mixins.ConversationViewMixin, CreateView):
    """View to start a new conversation."""
    pass


class ConversationRedirectView(view_mixins.ConversationViewMixin,
                               RedirectView):
    """View to list all conversations of a user."""
    permanent = False

    def get_redirect_url(self, **kwargs):
        if self.user.conversations.all():
            return reverse('conversation_update', kwargs={
                'pk': self.user.conversations.all()[0].pk})
        return reverse('conversation_create')


class ConversationArchiveView(view_mixins.ConversationStatusViewMixin,
                              DetailView):
    """View to archive a conversation."""
    action = 'archive'



"""Tests for the views of the ``conversation`` app."""
from django.contrib.contenttypes.models import ContentType
from django.core.urlresolvers import reverse
from django.test import TestCase

from django_libs.tests.factories import UserFactory
from django_libs.tests.views_tests import ViewTestMixin

from .factories import ConversationFactory


class ConversationRedirectViewTestCase(ViewTestMixin, TestCase):
    """Test case for the ConversationRedirectView view."""
    def setUp(self):
        self.user = UserFactory()

    def get_view_name(self):
        return 'conversation_list'

    def test_view(self):
        self.is_not_callable()
        self.is_callable(user=self.user,
                         and_redirects_to=reverse('conversation_create'))
        conversation = ConversationFactory()
        conversation.users.add(self.user)
        self.is_callable(and_redirects_to=reverse(
            'conversation_update', kwargs={'pk': conversation.pk}))


class ConversationCreateViewTestCase(ViewTestMixin, TestCase):
    """Test case for the ConversationCreateView view."""
    def setUp(self):
        self.user = UserFactory()
        self.stranger = UserFactory()

    def get_view_name(self):
        return 'conversation_create'

    def test_view(self):
        self.is_not_callable()
        self.is_callable(user=self.user)
        with self.settings(
                CONVERSATION_MESSAGE_FORM='conversation.forms.MessageForm'):
            self.is_callable(user=self.user)


class ConversationCreateViewInitialTestCase(ViewTestMixin, TestCase):
    """
    Test case for the ConversationCreateView view, if direct messaging a user.

    """
    def setUp(self):
        self.user = UserFactory()
        self.stranger = UserFactory()

    def get_view_name(self):
        return 'conversation_create_initial'

    def get_view_kwargs(self):
        return {'user_pk': self.stranger.pk}

    def test_view(self):
        self.is_not_callable()
        self.is_not_callable(user=self.user, kwargs={'user_pk': 999})
        self.is_callable()


class ConversationCreateViewContentObjectTestCase(ViewTestMixin, TestCase):
    """
    Test case for the ConversationCreateView view, if a content object is
    attached to the conversation.

    """
    def setUp(self):
        self.user = UserFactory()
        self.content_object = UserFactory()

    def get_view_name(self):
        return 'conversation_create_content_object'

    def get_view_kwargs(self):
        return {
            'c_type': ContentType.objects.get_for_model(self.content_object),
            'obj_id': self.content_object.pk,
        }

    def test_view(self):
        self.is_not_callable()
        new_kwargs = self.get_view_kwargs()
        new_kwargs.update({'obj_id': 999})
        self.is_not_callable(user=self.user, kwargs=new_kwargs)
        new_kwargs = self.get_view_kwargs()
        new_kwargs.update({'c_type': 'foobar'})
        self.is_not_callable(user=self.user, kwargs=new_kwargs)
        self.is_callable()


class ConversationCreateViewInitialContentObjectTestCase(ViewTestMixin,
                                                         TestCase):
    """
    Test case for the ConversationCreateView view, if a content object is
    attached to the conversation and if an initial user has been chosen.

    """
    def setUp(self):
        self.user = UserFactory()
        self.content_object = UserFactory()
        self.stranger = UserFactory()

    def get_view_name(self):
        return 'conversation_create_initial_content_object'

    def get_view_kwargs(self):
        return {
            'c_type': ContentType.objects.get_for_model(self.content_object),
            'obj_id': self.content_object.pk,
            'user_pk': self.stranger.pk,
        }

    def test_view(self):
        self.is_not_callable()
        self.is_callable(user=self.user)


class ConversationUpdateViewTestCase(ViewTestMixin, TestCase):
    """Test case for the ConversationUpdateView view."""
    def setUp(self):
        self.user = UserFactory()
        self.conversation = ConversationFactory()
        self.conversation.users.add(self.user)
        self.stranger = UserFactory()

    def get_view_name(self):
        return 'conversation_update'

    def get_view_kwargs(self):
        return {'pk': self.conversation.pk}

    def test_view(self):
        self.is_not_callable()
        self.is_not_callable(user=self.stranger)
        self.is_callable(user=self.user)
        self.is_callable(data={'text': 'Foobar'}, method='post')


class ConversationArchiveViewTestCase(ViewTestMixin, TestCase):
    """Test case for the ConversationArchiveView view."""
    def setUp(self):
        self.user = UserFactory()
        self.conversation = ConversationFactory()
        self.conversation.users.add(self.user)
        self.other_conversation = ConversationFactory()

    def get_view_name(self):
        return 'conversation_archive'

    def get_view_kwargs(self):
        return {'pk': self.conversation.pk}

    def test_view(self):
        self.is_not_callable()
        self.is_not_callable(user=self.user)
        self.is_not_callable(kwargs={'pk': self.other_conversation.pk},
                             method='post', ajax=True)
        self.is_callable(method='post', ajax=True)
        self.assertIn(
            self.user, self.conversation.archived_by.all(),
            msg=('The conversation should have been marked as archived.'))
